"""
general HTML generation utilities
"""

# types
from ast import AST
from typing  import Tuple, Generator, List

HTMLGenerator = Generator[str, None, None] # yield str, receive None from send(), return None

HTML = Tuple[HTMLGenerator]
Classes = str | List[str]


# decorator for HTML generators
# usage:
# @debug
# def render_...
def debug(f):
    def g(node):
        for x in f(node):
            print(x)
            yield x
    return g




"""
(bidirectional) AST-DOM linking

AST nodes are linked to DOM elements with a shared unique id
"""
# TODO: centralize IDs
from untext.rendering.dynamic.dom import genid, register, ast_mapping

#def register(ast_node: AST, dom_element: Element):
#    n = genid()
#    dom_element.id = n
#    # .id is used by some node types
#    ast_node.node_id = n
#    dom_mapping[n] = dom_element
#    ast_mapping[n] = ast_node

"""
basic wrappers for html string formatting

HTML output is generated in ~O(n) where n is the size of the html code.
This is possible because of generators.
Expect lots of yield statements.
"""

# helpers
def id_attr(id) -> HTMLGenerator:
    # " id='{id}'"
    yield " id='"
    yield str(id)
    yield "'"

def class_attr(classes: Classes) -> HTMLGenerator:
    # " class='{...classes}'"
    yield " class='"
    if isinstance(classes, str):
        yield classes
    else:
        yield " ".join(classes)
    yield "'"

def data_attr(attr: dict) -> HTMLGenerator:
    # " data-...='...'"
    for key, val in attr.items():
        yield " data-"
        yield key
        yield "='"
        yield val
        yield "'"


# main way of generating html
# we use a div soup because writing wrappers for every
# html tag adds complexity for no benefit
def div(*items: HTML, id=None, classes: Classes = [], attr: dict = {}) -> HTMLGenerator:
    # <div {id} {class}>{items}</div>
    yield "<div"
    if id is not None:
        yield from id_attr(id)
    if classes:
        yield from class_attr(classes)
    if data_attr:
        yield from data_attr(attr)
    yield ">"
    for x in items:
        yield from x
    yield "</div>"


def node(n: AST, *items: HTML) -> HTMLGenerator:
    id = genid()
    n.node_id = id
    # dom_mapping is useless because the Element does not exist yet and will have an id anyway
    ast_mapping[id] = n
    yield from div(*items, id=id)


# usage: div(text("hello"))
def text(x: str) -> HTMLGenerator:
    yield x


# TODO: name
def element(classes: Classes, *items: HTML) -> HTMLGenerator:
    yield from div(classes=classes, *items)


# def block(html=""):
#     return f"<div class='block'>{html}</div>"
#
# def add_pre(parent: Element, text: str):
#     elt = parent.append(pre())
#     elt.text = text
#     return elt

# TODO: find an API to add children with wrapping divs

#def add_managed_item(parent: Element, )
    #def comma_separated_list(parent: Element):
    #    elt = add(parent, "comma-sep")
    #    return elt
    #
    #def add_item(lst, *args)

#def add(parent, html):
#    parent.append(html)

#def row():
#    return "<div class='row'></div>"


"""
high level helpers
"""

# decorator for (node: AST) -> HTMLGenerator html renderers
def register_node(f):
    def renderer(n: AST):
        yield from node(n, f(n))
    return renderer

def items(parent_style, item_style, items):
    items = [element(item_style, item) for item in items]
    parent = element(parent_style, *items)
    return parent






"""
expression rendering

from the official python documentation:
https://docs.python.org/3/library/ast.html
(at the time of writing, for python 3.14)


expr = BoolOp(boolop op, expr* values)
    | NamedExpr(expr target, expr value)
    | BinOp(expr left, operator op, expr right)
    | UnaryOp(unaryop op, expr operand)
    | Lambda(arguments args, expr body)
    | IfExp(expr test, expr body, expr orelse)
    | Dict(expr?* keys, expr* values)
    | Set(expr* elts)
    | ListComp(expr elt, comprehension* generators)
    | SetComp(expr elt, comprehension* generators)
    | DictComp(expr key, expr value, comprehension* generators)
    | GeneratorExp(expr elt, comprehension* generators)
    -- the grammar constrains where yield expressions can occur
    | Await(expr value)
    | Yield(expr? value)
    | YieldFrom(expr value)
    -- need sequences for compare to distinguish between
    -- x < 4 < 3 and (x < 4) < 3
    | Compare(expr left, cmpop* ops, expr* comparators)
    | Call(expr func, expr* args, keyword* keywords)
    | FormattedValue(expr value, int conversion, expr? format_spec)
    | Interpolation(expr value, constant str, int conversion, expr? format_spec)
    | JoinedStr(expr* values)
    | TemplateStr(expr* values)
    | Constant(constant value, string? kind)

    -- the following expression can appear in assignment context
    | Attribute(expr value, identifier attr, expr_context ctx)
    | Subscript(expr value, expr slice, expr_context ctx)
    | Starred(expr value, expr_context ctx)
    | Name(identifier id, expr_context ctx)
    | List(expr* elts, expr_context ctx)
    | Tuple(expr* elts, expr_context ctx)

    -- can appear only in Subscript
    | Slice(expr? lower, expr? upper, expr? step)

    -- col_offset is the byte offset in the utf8 string the parser uses
    attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
"""


from webview.dom.element import Element
import ast

import html as pyhtml
import json

#from untext.rendering.html import register, div, add_node, add, add_text, add_pre
from .html import text, element, debug, register_node, div
from . import html


def render(node: ast.expr):
    #match type(node):
    if type(node) == ast.BoolOp:
        yield from render_boolop(node)
    elif type(node) == ast.NamedExpr:
        raise NotImplementedError('expression.render() not implemented for ast.NamedExpr')
    elif type(node) == ast.BinOp:
        yield from render_binop(node)
    elif type(node) == ast.UnaryOp:
        yield from render_unaryop(node)
    elif type(node) == ast.Lambda:
        raise NotImplementedError('expression.render() not implemented for ast.Lambda')
    elif type(node) == ast.IfExp:
        yield from render_ifexp(node)
    elif type(node) == ast.Dict:
        yield from render_dict(node)
    elif type(node) == ast.Set:
        raise NotImplementedError('expression.render() not implemented for ast.Set')
    elif type(node) == ast.ListComp:
        yield from render_list_comprehension(node)
    elif type(node) == ast.SetComp:
        raise NotImplementedError('expression.render() not implemented for ast.SetComp')
    elif type(node) == ast.DictComp:
        raise NotImplementedError('expression.render() not implemented for ast.DictComp')
    elif type(node) == ast.GeneratorExp:
        raise NotImplementedError('expression.render() not implemented for ast.GeneratorExp')
    elif type(node) == ast.Await:
        raise NotImplementedError('expression.render() not implemented for ast.Await')
    elif type(node) == ast.Yield:
        yield from render_yield(node)
    elif type(node) == ast.YieldFrom:
        yield from render_yieldfrom(node)
    elif type(node) == ast.Compare:
        yield from render_compare(node)
    elif type(node) == ast.Call:
        yield from render_call(node)
    elif type(node) == ast.FormattedValue:
        yield from render_formatted_value(node)
    # case ast.Interpolation:
    #     raise NotImplementedError('expression.render() not implemented for ast.Interpolation')
    elif type(node) == ast.JoinedStr:
        yield from render_joinedstr(node)
    # case ast.TemplateStr:
    #     raise NotImplementedError('expression.render() not implemented for ast.TemplateStr')
    elif type(node) == ast.Constant:
        yield from render_constant(node)
    elif type(node) == ast.Attribute:
        yield from render_attribute(node)
    elif type(node) == ast.Subscript:
        yield from render_subscript(node)
    elif type(node) == ast.Starred:
        yield from render_starred(node)
    elif type(node) == ast.Name:
        yield from render_name(node)
    elif type(node) == ast.List:
        yield from render_list(node)
    elif type(node) == ast.Tuple:
        yield from render_tuple(node)
    elif type(node) == ast.Slice:
        yield from render_slice(node)
    else:
        raise ValueError(f"Unexpected ast expression type: {type(node)}")


"""
AST expression rendering

(implementation for each type)
"""


@register_node
def render_boolop(node: ast.BoolOp):
    values = [render(v) for v in node.values]
    result = html.items(f"row gap {read_boolop(node.op)}-sep", "row gap", values)
    yield from result


# part of render_boolop
def read_boolop(op: ast.boolop):
    if isinstance(op, ast.And):
        return "and"
    elif isinstance(op, ast.Or):
        return "or"
    else:
        raise NotImplementedError(f"unknown boolean operator: {op}")


"""
this code uses html data-attributes to encode the operator, and css to render it dynamically

the html output in theory:
<div class="operation" data-operator="+">
  <div class="operand">a</div>
  <div class="operand">b</div>
</div>

in practice, css support is lacking, so we need this:
<div class="operation" data-operator="+">
  <div class="operand" data-operator="+">a<div>
  <div class="operand">b</div>
</div>
"""
@register_node
def render_binop(node: ast.BinOp):
    left = render(node.left)
    right = render(node.right)
    # TODO: find better abstractions
    # currently, using data attributes can only be done with div()
    right = div(right,
                classes="row gap",
                attr={"operator": read_binaryop(node.op)})

    result = element("operation row gap", left, right)
    yield from result


def read_binaryop(op: ast.operator):
    # css classes cannot have special characters like +
    if isinstance(op, ast.Add):
        return "+"
    elif isinstance(op, ast.Sub):
        return "-"
    elif isinstance(op, ast.Mult):
        return "*"
    elif isinstance(op, ast.Div):
        pass
    elif isinstance(op, ast.FloorDiv):
        pass
    elif isinstance(op, ast.Mod):
        pass
    elif isinstance(op, ast.Pow):
        pass
    elif isinstance(op, ast.LShift):
        pass
    elif isinstance(op, ast.RShift):
        pass
    elif isinstance(op, ast.BitOr):
        return "|"
    elif isinstance(op, ast.BitXor):
        pass
    elif isinstance(op, ast.BitAnd):
        pass
    elif isinstance(op, ast.MatMult):
        pass
    raise NotImplementedError(f"unknown binary operator: {op}")


# TODO: go back to previous "operator separators" and replace them by DOM nodes
# (operators have semantic meaning, they are not just syntax)
# (keep the css for infix inlining if needed)
def render_unaryop(node: ast.UnaryOp):
    value = render(node.operand)
    op = read_unaryop(node.op)
    yield from div(value,
                   classes="unary-operation row gap",
                   attr={"operator": op})


def read_unaryop(op: ast.unaryop):
    if isinstance(op, ast.USub):
        return "-"
    elif isinstance(op, ast.UAdd):
        return "+"
    elif isinstance(op, ast.Not):
        return "not "
    elif isinstance(op, ast.Invert):
        return "~"
    else:
        raise NotImplementedError("unknown unary operator")


@register_node
def render_ifexp(node: ast.IfExp):
    test = render(node.test)

    if_expr = render(node.body)
    if_part = element("if-prefix row gap", if_expr)

    else_expr = render(node.orelse)
    else_part = element("else-prefix row gap", else_expr)

    if_expr = element("row gap", test, if_part, else_part)
    yield from element("if-expression", if_expr)


# TODO: test with more kinds of literals than "{}"
def render_dict(node: ast.Dict):
    # TODO: multi-line rendering:
    # ... {
    #     ....
    # }
    # for now: render on a single line
    rows = []
    for i in range(len(node.keys)):
        k = node.keys[i]
        # TODO: None is for **d unpacking
        assert k is not None
        v = node.values[i]
        key = render(k)
        val = render(v)

        key = element("row colon-suffix", key)
        val = element("", val)

        row = element("row gap", key, val)
        rows.append(row)

    items = html.items("comma-sep", "", rows)
    result = element("row braces", items)
    yield from element("dict", result)


@register_node
def render_list_comprehension(node: ast.ListComp):
    # TODO: test with more than 1 generator
    generators = []
    for g in node.generators:
        generators.append(render_comprehension_generator(g))
    elt = render(node.elt)
    generators = element("row gap", *generators)
    content = element("row gap", elt, generators)
    yield from element("list-comprehension brackets row", content)


@register_node
def render_comprehension_generator(node: ast.comprehension):
    # TODO: support async
    assert node.is_async == 0

    conds = []
    for cond in node.ifs:
        cond = render(cond)
        conds.append(cond)
    conds = html.items("row gap", "if-prefix row gap", conds)

    target = render(node.target)
    iterated = render(node.iter)
    generator = html.items("in-sep row gap", "row gap", [target, iterated])
    yield from element("comprehension-generator for-prefix row gap", generator)



@register_node
def render_yield(node):
    expr = render(node.value)
    yield from element("yield yield-prefix row gap", expr)


@register_node
def render_yieldfrom(node):
    expr = render(node.value)
    yield from element("yield-from yield-from-prefix row gap", expr)


def render_compare(node: ast.Compare):
    # in python, comparisons can be complex sequences, like:
    # 1 < x < y < 6
    # 1 is called left
    # the operators are [<, <, <]
    # and the comparators are [x, y, 6]
    left = render(node.left)
    elts = [left]
    for op, cmp in zip(node.ops, node.comparators):
        op = text(read_op(op))
        cmp = render(cmp)
        elts.append(op)
        elts.append(cmp)
    yield from element("compare row gap", *elts)


def read_op(op: ast.operator):
    if isinstance(op, ast.Eq):
        return "=="
    elif isinstance(op, ast.NotEq):
        return "!="
    elif isinstance(op, ast.Lt):
        return "<"
    elif isinstance(op, ast.LtE):
        return "<="
    elif isinstance(op, ast.Gt):
        return ">"
    elif isinstance(op, ast.GtE):
        return ">="
    elif isinstance(op, ast.Is):
        return "is"
    elif isinstance(op, ast.IsNot):
        return "is not"
    elif isinstance(op, ast.In):
        return "in"
    elif isinstance(op, ast.NotIn):
        return "not in"
    else:
        raise NotImplementedError("unknown comparison operator")




# TODO: add more DOM encoding
@register_node
def render_call(node: ast.Call):
    func = render(node.func)
    args = [render(arg) for arg in node.args]
    kwargs = [render_keyword_arg(kwarg) for kwarg in node.keywords]
    args = html.items("comma-sep row", "row gap", args + kwargs)
    args = element("parens row", args)
    yield from element("call row", func, args)


# part of render_call(), also used by statement.render_class()
@register_node
def render_keyword_arg(node: ast.keyword):
    kw = text(node.arg)
    val = render(node.value)
    yield from html.items("keyword-argument equal-sep row", "row", [kw, val])


@register_node
def render_formatted_value(node: ast.FormattedValue):
    # TODO: support other conversion types:
    # -1: unspecified (default is str())
    # 97: !a, ascii
    # 114: !r, repr() output
    # 115: !s, str() output
    assert node.conversion == -1
    # can be a nested JoinedStr instead of None
    # TODO: support this attribute
    assert node.format_spec is None
    expr = render(node.value)
    yield from element("f-value", expr)


# f"{x}<text>{y}"
# f-"-({(<x>)}-(<json-encoded text>)-({<y>}))-"
# TODO: remove unneeded .string-literal classes
@register_node
def render_joinedstr(node: ast.JoinedStr):
    parts = []
    for e in node.values:
        if isinstance(e, ast.FormattedValue):
            e = render(e)
            braced = element("braces row", e)
            string_styled = element("string-literal", braced)
            parts.append(string_styled)
        else:
            assert isinstance(e, ast.Constant)
            assert isinstance(e.value, str)
            str_text = html_serialize_str(e.value)
            # remove quotes (and let the css quotes surround the whole f-string)
            # quotes are converted to "&quot;" by html_serialize_str
            str_text = str_text[6:-6]
            # TODO: see if string-literal is better for the whole f-string
            txt = text(str_text)
            txt = element("string-literal", txt)
            parts.append(txt)

    quoted = element("quotes row", *parts)
    string_styled = element("string-literal", quoted)
    yield from element("row f-prefix", string_styled)


@register_node
def render_constant(node: ast.Constant):
    assert node.kind is None

    if isinstance(node.value, str):
        # TODO: test with multiline strings,
        # make it work with the DOM renderer,
        # and refactor into a function
        str_text = html_serialize_str(node.value)
        txt = element("string-literal", text(str_text))
        const = div(txt, classes="literal", attr={"const-type": "str"})
    else:
        typename = type(node.value).__name__
        value_text = text(repr(node.value))
        const = div(value_text,
                  classes="literal",
                  attr={"const-type": typename})
    yield from element("constant", const)


# used by render_constant and render_joinedstr
def html_serialize_str(txt: str) -> str:
    return pyhtml.escape(json.dumps(txt))


@register_node
def render_attribute(node: ast.Attribute):
    obj = render(node.value)
    attr = text(node.attr)
    row = html.items("attribute row dot-sep", "row", [obj, attr])
    yield from row


@register_node
def render_subscript(node: ast.Subscript):
    # node.ctx is either ast.Load or ast.Store
    # Store if the subscript is in a left side of an assignment
    # Load if the subscript is in an expression to evaluate
    index = render(node.slice)
    bracketed = element("brackets row", index)
    indexed = render(node.value)
    yield from element("subscript row", indexed, bracketed)


@register_node
def render_starred(node: ast.Starred):
    expr = render(node.value)
    yield from element("starred star-prefix row", expr)


@register_node
def render_name(node: ast.Name):
    yield from element("symbol", text(node.id))


@register_node
def render_list(node: ast.List):
    assert isinstance(node.ctx, ast.Load)
    elts = [render(x) for x in node.elts]
    elts = html.items("comma-sep row", "row gap", elts)
    yield from element("list brackets row", elts)


@register_node
def render_tuple(node: ast.Tuple):
    #print(node.ctx)
    # TODO: cleanup this branching
    if len(node.elts) == 0:
        yield from element("tuple parens row")
        return
    if len(node.elts) == 1:
        # display a single item with a comma after it
        expr = render(node.elts[0])
        # empty element for the comma
        empty_elt = element()
        comma_separated = html.items("comma-sep row", "row", [expr, empty_elt])
        yield from element("tuple parens row", comma_separated)
        return
    else:
        elts = [render(e) for e in node.elts]
        elts = html.items("comma-sep row", "row gap", elts)
        yield from element("tuple parens row", elts)
        return


@register_node
def render_slice(node: ast.Slice):
    # a slice must have a left and right part separated by :, even if they are implicit
    parts = [element("optional-slot"), element("optional-slot")]
    if node.lower is not None:
        parts[0] = render(node.lower)
    if node.upper is not None:
        parts[1] = render(node.upper)
    if node.step is not None:
        parts.append(render(node.step))
    yield from html.items("slice row colon-sep", "row", parts)




"""
statement rendering

from the official python AST documentation:
https://docs.python.org/3/library/ast.html
(at the time of writing, for python 3.14)


stmt = FunctionDef(identifier name, arguments args,
                   stmt* body, expr* decorator_list, expr? returns,
                   string? type_comment, type_param* type_params)
      | AsyncFunctionDef(identifier name, arguments args,
                         stmt* body, expr* decorator_list, expr? returns,
                         string? type_comment, type_param* type_params)

      | ClassDef(identifier name,
         expr* bases,
         keyword* keywords,
         stmt* body,
         expr* decorator_list,
         type_param* type_params)
      | Return(expr? value)

      | Delete(expr* targets)
      | Assign(expr* targets, expr value, string? type_comment)
      | TypeAlias(expr name, type_param* type_params, expr value)
      | AugAssign(expr target, operator op, expr value)
      -- 'simple' indicates that we annotate simple name without parens
      | AnnAssign(expr target, expr annotation, expr? value, int simple)

      -- use 'orelse' because else is a keyword in target languages
      | For(expr target, expr iter, stmt* body, stmt* orelse, string? type_comment)
      | AsyncFor(expr target, expr iter, stmt* body, stmt* orelse, string? type_comment)
      | While(expr test, stmt* body, stmt* orelse)
      | If(expr test, stmt* body, stmt* orelse)
      | With(withitem* items, stmt* body, string? type_comment)
      | AsyncWith(withitem* items, stmt* body, string? type_comment)

      | Match(expr subject, match_case* cases)

      | Raise(expr? exc, expr? cause)
      | Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
      | TryStar(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
      | Assert(expr test, expr? msg)

      | Import(alias* names)
      | ImportFrom(identifier? module, alias* names, int? level)

      | Global(identifier* names)
      | Nonlocal(identifier* names)
      | Expr(expr value)
      | Pass | Break | Continue

      -- col_offset is the byte offset in the utf8 string the parser uses
      attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
"""


# types
# HTML renderer -> no DOM needed
#from webview.dom.dom import DOM
#from webview.dom.element import Element
import ast

# html generation wrappers
from .html import node, text, element, debug, register_node
from . import html

# expressions can be found inside statements, but not the opposite
# (statement.py imports expression.py, but not the opposite)
from . import expression



"""
AST module rendering

(modules are not technically statements)
"""


def render_module(node: ast.Module):
    # unknown value
    assert len(node.type_ignores) == 0
    items = []
    for elt in node.body:
        # top-level strings are usually multiline
        # TODO: refactor to make the logic more explicit
        if isinstance(elt, ast.Expr) and isinstance(elt.value, ast.Constant) and isinstance(elt.value.value, str):
            lines = elt.value.value.split("\n")
            lines[0] = '"""' + lines[0]
            lines[-1] += '"""'
            lines = [line if line else "<br>" for line in lines]
            lines = [f"<div>{line}</div>" for line in lines]
            #items.append("".join(lines))
            items.append(html.text("".join(lines)))
            # failed implementation
            #rendered_text = elt.value.value.replace("\n", "</div><div>")
            #print(rendered_text)
            #print("\n" in rendered_text)
            #children.append(f'<div>"""</div><div>{elt.value.value}</div><div>"""</div>')
        else:
            items.append(render(elt))
    yield from element("module", *items)



"""
AST statement rendering

(the big switch)
"""


def render(node: ast.stmt):
    #match type(node):
    if type(node) == ast.FunctionDef:
        yield from render_funcdef(node)
    elif type(node) == ast.AsyncFunctionDef:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")

    elif type(node) == ast.ClassDef:
        yield from render_classdef(node)
    elif type(node) == ast.Return:
        yield from render_return(node)

    elif type(node) == ast.Delete:
        yield from render_delete(node)
    elif type(node) == ast.Assign:
        yield from render_assign(node)
    # 3.12+ feature
    # TODO: ignore it until pypy reaches 3.12 or stop supporting pypy
    #case ast.TypeAlias:
    #    raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.AugAssign:
        yield from render_augassign(node)
    elif type(node) == ast.AnnAssign:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")

    elif type(node) == ast.For:
        yield from render_for(node)
    elif type(node) == ast.AsyncFor:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.While:
        yield from render_while(node)
    elif type(node) == ast.If:
        yield from render_if(node)
    elif type(node) == ast.With:
        yield from render_with(node)
    elif type(node) == ast.AsyncWith:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")

    elif type(node) == ast.Match:
        yield from render_match(node)

    elif type(node) == ast.Raise:
        yield from render_raise(node)
    elif type(node) == ast.Try:
        # TODO: implement
        #yield from render_try(node)
        raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.TryStar:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.Assert:
        yield from render_assert(node)

    elif type(node) == ast.Import:
        yield from render_import(node)
    elif type(node) == ast.ImportFrom:
        yield from render_importfrom(node)

    elif type(node) == ast.Global:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.Nonlocal:
        yield from render_nonlocal(node)
    elif type(node) == ast.Expr:
        # TODO: add a wrapper div, to type as a "expr in a statement"
        yield from expression.render(node.value)
    elif type(node) == ast.Pass:
        yield from render_pass(node)
    elif type(node) == ast.Break:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")
    elif type(node) == ast.Continue:
        yield from render_continue(node)

    # future python versions may add new things
    else:
        raise ValueError(f"Unexpected ast statement type: {type(node)}")



"""
AST statement rendering

(implementation for each type)
"""


@register_node
def render_funcdef(node: ast.FunctionDef):
    # supported features checks
    #assert len(node.decorator_list) == 0
    assert node.type_comment is None
    # 3.12+ feature
    # instead, see: type_comment
    # TODO: wait for pypy to reach 3.12 or explicitely stop supporting pypy
    #assert len(node.type_params) == 0

    # decorators
    decorators = [expression.render(d) for d in node.decorator_list]
    decorators = [element("at-prefix row", d) for d in decorators]
    # TODO: decide if subcomponents (decorator list, header, body)
    # should be annotated with classes or not
    decorators = element("decorators", *decorators)

    # header
    name = text(node.name)
    funcname = element("row def-prefix gap", name)
    params = render_parameters(node.args)
    funcparams = element("parens row", params)
    head = element("row", funcname, funcparams)
    if node.returns is not None:
        # "def f(...)" -> "def f(...) -> ..."
        funcreturn = expression.render(node.returns)
        head = html.items("row gap return-type-arrow-sep", "row gap",
                          [head, funcreturn])
    header = element("row colon-suffix", head)

    # body
    body = [render(stmt) for stmt in node.body]
    body_block = element("block", *body)

    yield from element("def", decorators, header, body_block)


# sub-part of render_funcdef
@register_node
def render_parameters(node: ast.arguments):
    # assertions and attribute parsing
    # TODO: support more cases
    assert len(node.posonlyargs) == 0
    # default values are for the last parameters
    # need to match argument names to default values with indices
    default_padding = len(node.args) - len(node.defaults)
    # flags (*args and **kwargs)
    assert node.kwarg is None

    params = []
    # normal args
    for i, param in enumerate(node.args):
        param = render_param(param)
        if i >= default_padding:
            # TODO: refactor the <param> = <value> rendering into render_param()
            default_value = node.defaults[i - default_padding]
            default_value = expression.render(default_value)
            param = html.items(
                "equal-sep row gap",
                "row gap",
                [param, default_value]
            )
        params.append(param)

    # vararg
    if node.vararg is not None:
        param = render_param(node.vararg)
        starred = element("star-prefix row", param)
        params.append(starred)

    # kwonly args (args that cannot be positional due to the vararg before them)

    # node.kwonlyargs are the parameters after the *vararg
    # the *vararg comes after normal (non-kwonly) parameters
    # their default values are in kw_defaults

    # node.kw_default contains default values for every kwonly arg,
    # with None when there is no default value
    for i, param in enumerate(node.kwonlyargs):
        param = render_param(param)
        if node.kw_defaults[i] is not None:
            default_value = node.kw_defaults[i]
            # items of a list with a separator need a wrapper to style the separator
            default_value = expression.render(default_value)
            param = html.items(
                "equal-sep row gap",
                "row gap",
                [param, default_value]
            )
        params.append(param)


    yield from html.items("parameters comma-sep row", "row gap", params)


# sub-part of render_parameters
@register_node
def render_param(node: ast.arg):
    assert node.type_comment is None

    param_name = text(node.arg)
    # used to indicate the structure of the html:
    param_name = element("parameter-name", param_name)
    if node.annotation is None:
        #param_name = element("bg-red", param_name)
        yield from param_name
        return

    annotation = expression.render(node.annotation)
    name = element("row colon-suffix", param_name)
    param = element("parameter row gap", name, annotation)
    yield from param


def render_classdef(node: ast.ClassDef):
    # TODO: support decorators
    # TODO: support type_params
    assert not node.decorator_list

    # decorators
    # TODO:
    decorators = element("")

    # header
    name = text(node.name)
    kwargs = [expression.render(kwarg) for kwarg in node.keywords]
    args = [expression.render(arg) for arg in node.bases]
    if kwargs or args:
        arguments = html.items("comma-sep row", "row gap", [*kwargs, *args])
        arguments = element("parens row", arguments)
    else:
        arguments = element("")
    header = element("row", name, arguments)
    header = element("class-prefix row gap", header)
    header = element("row colon-suffix", header)

    # body
    body = [render(stmt) for stmt in node.body]
    body = element("block", *body)

    yield from element("class", decorators, header, body)


@register_node
def render_return(node: ast.Return):
    if node.value is not None:
        value = expression.render(node.value)
    else:
        value = text("")
    yield from element("return return-prefix row gap", value)


@register_node
def render_delete(node: ast.Delete):
    # example: del a, b, c
    items = [expression.render(target) for target in node.targets]
    items = html.items("comma-sep row", "row gap bg-red", items)
    yield from element("delete del-prefix row gap bg-red", items)


@register_node
def render_assign(node: ast.Assign):
    assert node.type_comment is None

    value = expression.render(node.value)
    targets = [expression.render(t) for t in node.targets]
    # display the value as the last (equal-separated) target
    targets.append(value)
    formatted = html.items("assign equal-sep row gap", "row gap", targets)
    yield from formatted



# augassign:
#targ = render_expr(node.target)
#op = render_binaryop(node.op)
#val = render_expr(node.value)
#return f"{targ} {op}= {val}"

# TODO: fix: the operator must be displayed
# format: <node> <op>= <node>
# <node><gap><op>=<gap><node>
# (<node> (<op>=) <node>)
def render_augassign(node: ast.AugAssign):
    target = expression.render(node.target)
    operator = text(expression.read_binaryop(node.op))
    # hack: add an empty div to force the = separator to render
    # TODO?: add a .equal-suffix css class
    empty = element("")
    operator = html.items("equal-sep row", "row", [operator, empty])
    val = expression.render(node.value)
    yield from element("row gap", target, operator, val)


@register_node
def render_for(node: ast.For):
    assert node.type_comment is None

    # header
    # example: "for x in lst:"
    variable = expression.render(node.target)  # "x"
    iterator = expression.render(node.iter)  # "lst"
    # "x in lst"
    # TODO: could fit .for-prefix here
    header = html.items("in-sep row gap", "row gap",
                                [variable, iterator])
    # "for x in lst"
    header = element("for-prefix row gap", header)
    # "for x in lst:"
    header = element("colon-suffix row", header)

    body = [render(stmt) for stmt in node.body]
    body = element("block", *body)


    # TODO: process for: else: blocks
    # (add more headers after the main body)
    assert not node.orelse
    #parts = [block]
    #if node.orelse:
    #    else_header = div("else:")
    #    else_body = [render_statement(statement) for statement in node.orelse]
    #    else_body = block("".join(else_body))
    #    else_block = div(else_header + else_body)
    #    parts.append(else_block)

    #result = "".join(parts)
    #return div(result)

    yield from element("for", header, body)



@register_node
def render_while(node: ast.While):
    test = expression.render(node.test)
    header = element("while-prefix row gap", test)
    header = element("colon-suffix row", header)

    body = [render(stmt) for stmt in node.body]
    body = element("block", *body)

    assert not node.orelse
    # if node.orelse:
    #else_body = [render(stmt) for stmt in node.orelse]
    yield from element("while", header, body)


@register_node
def render_if(node: ast.If):
    if is_elif(node):
        yield from render_elifs(node)
        return

    test = expression.render(node.test)
    header = element("row gap if-prefix", test)
    header = element("row colon-suffix", header)

    if_body = [render(stmt) for stmt in node.body]
    if_block = element("block", *if_body)

    parts = [header, if_block]

    if node.orelse:
        else_header = element("row colon-suffix else-prefix")
        else_body = [render(stmt) for stmt in node.orelse]
        else_block = element("block", *else_body)
        parts.append(else_header)
        parts.append(else_block)
    yield from element("if", *parts)


# helpers for elifs
# internally, Python represents elifs as nested else: if:
def is_elif(if_node: ast.If):
    "return True if the if has an elif branch"
    # if/elif gets compiled to if/else{if}
    # if/elif/else gets compiled to if/else{if/else}
    if len(if_node.orelse) != 1:
        return False
    return isinstance(if_node.orelse[0], ast.If)

def elif_depth(if_node: ast.If):
    "return the number of elif branches in an if"
    n = 0
    while is_elif(if_node):
        n += 1
        if_node = if_node.orelse[0]
    return n

def collect_branches(if_node: ast.If):
    n = elif_depth(if_node)
    branches = []
    for i in range(n):
        condition = if_node.test
        body = if_node.body
        branches.append((condition, body))
        if_node = if_node.orelse[0]
    # final else (not if) branch
    if if_node.orelse:
        assert not isinstance(if_node.orelse[0], ast.If)
    # the last element is not a (cond, body) tuple
    branches.append(if_node.orelse)
    return branches


# the node registering is already done by render_if()
def render_elifs(if_node: ast.If):
    branches = collect_branches(if_node)
    if_branch, *elif_branches, else_body = branches


    blocks = []

    # if
    test, body = if_branch
    test = expression.render(test)
    header = element("row gap if-prefix", test)
    header = element("row colon-suffix", header)

    body = [render(stmt) for stmt in body]
    body = element("block", *body)

    if_block = element("if-block", header, body)
    blocks.append(if_block)

    # elifs
    for (test, body) in elif_branches:
        test = expression.render(test)
        header = element("row gap elif-prefix", test)
        header = element("row colon-suffix", header)

        body = [render(stmt) for stmt in body]
        body = element("block", *body)

        elif_block = element("elif-block", header, body)
        blocks.append(elif_block)

    # else
    if else_body:
        header = element("row else-prefix colon-suffix")
        body = [render(stmt) for stmt in else_body]
        body = element("block", *body)
        else_block = element("else-block", header, body)
        blocks.append(else_block)
        #header = add(else_elt, "row colon-suffix")
        #header_content = add(header, "row gap else-prefix")

    # TODO: see if render_if() and render_elif() can return the same html layou
    # if possible, use the same .if class for both
    yield from element("elif", *blocks)


@register_node
def render_with(node: ast.With):
    assert node.type_comment is None
    assert len(node.items) == 1  # TODO: test with more cases
    items = [render_withitem(item) for item in node.items]
    body = [render(stmt) for stmt in node.body]
    block = element("block", *body)
    header = element("with-prefix row gap", *items)
    header = element("row colon-suffix", header)
    yield from element("with", header, block)


@register_node
def render_withitem(node: ast.withitem):
    if node.optional_vars:
        # "<expr> as <name>"
        expr = expression.render(node.context_expr)
        name = expression.render(node.optional_vars)
        item = html.items("as-sep row gap", "row gap", [expr, name])
    else:
        # just "<expr>"
        item = expression.render(node.context_expr)
    yield from element("with-item", item)


# TODO: write tests to make sure the feature does not bitrot
# (cython prevents the untext codebase to use match blocks)
@register_node
def render_match(node: ast.Match):
    matched = expression.render(node.subject)
    header = element("match-prefix row gap", matched)
    header = element("colon-suffix row", header)

    cases = [render_case(case) for case in node.cases]
    cases = element("block", *cases)

    yield from element("match", header, cases)


# part of render_match
@register_node
def render_case(node: ast.match_case):
    # TODO: support more cases
    assert node.guard is None
    body = [render(stmt) for stmt in node.body]
    body = element("block", *body)
    # turns out match patterns are not actually expressions,
    # despite having the same syntax
    pattern = render_pattern(node.pattern)
    pattern = element("row gap case-prefix", pattern)
    pattern = element("row colon-suffix", pattern)
    yield from element("case", pattern, body)


# TODO: implement missing cases
@register_node
def render_pattern(node: ast.pattern):
    #match type(node):
    if type(node) == ast.MatchValue:
        yield from render_match_value(node)
    elif type(node) == ast.MatchSingleton:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_singleton(node)
    elif type(node) == ast.MatchSequence:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_sequence(node)
    elif type(node) == ast.MatchMapping:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_mapping(node)
    elif type(node) == ast.MatchClass:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_class(node)

    elif type(node) == ast.MatchStar:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_star(node)

    elif type(node) == ast.MatchAs:
        yield from render_match_as(node)
    elif type(node) == ast.MatchOr:
        raise NotImplementedError(f"Unknown match pattern type: {type(node)}")
        #yield from render_match_or(node)

    else:
        raise ValueError(f"Unexpected match pattern type: {type(node)}")


@register_node
def render_match_value(node: ast.MatchValue):
    expr = expression.render(node.value)
    yield from element("match-value", expr)


@register_node
def render_match_as(node: ast.MatchAs):
    # TODO: support other cases
    # case [x] as y:
    # case default:     <--- the only kind used

    # specific to case default:
    assert node.pattern is None
    assert node.name == "default"
    txt = text(node.name)
    yield from element("match-as", txt)


@register_node
def render_raise(node: ast.Raise):
    # TODO: check if support for this attribute is needed
    assert node.cause is None
    raised = expression.render(node.exc)
    yield from element("raise raise-prefix row gap", raised)


@register_node
def render_assert(node: ast.Assert):
    # TODO: support assertion messages
    assert node.msg is None
    asserted = expression.render(node.test)
    yield from element("assert assert-prefix row gap", asserted)


# TODO: rewrite
def render_import(node: ast.Import):
    aliases = []
    for name in node.names:
        continue
        aliases.append(
            element(
                "row gap",
                render_alias(name)
            )
        )
    yield from html.node(
        node,
        element(
            "import import-prefix row",
            element(
                "aliases row comma-sep",
                #*aliases,
                *[element("row gap", render_alias(name))
                    for name in node.names]
            )
        )
    )

# sub-part of import and importfrom nodes
def render_alias(node: ast.alias):
    if node.asname is not None:
        alias = html.element(
            "named-alias import-alias as-sep row gap",
            html.element("row gap", html.text(node.name)),
            html.element("row gap", html.text(node.asname))
        )
    else:
        alias = html.element(
            "unnamed-alias",
            html.text(node.name)
        )
    yield from html.node(
        node,
        html.element(
            "alias row",
            alias
        )
    )


def render_importfrom(node: ast.ImportFrom):
    # build the html bottom-up
    # example: from ..a import b, c as d

    # ".."
    import_level = html.text("." * node.level)
    # "a"
    from_module = html.text(node.module or "")
    # ["b", "c as d"]
    imported = [render_alias(name) for name in node.names]
    # (comma separated) list items must be wrapped for styling separators
    imported = [html.element("row gap", x) for x in imported]

    # "b, c as d"
    imported_field = html.element("aliases row comma-sep", *imported)
    # "..a"
    from_field = html.element("row", import_level, from_module)

    # "import b, c as d"
    import_part = html.element("import-prefix row gap", imported_field)
    # "from ..a"
    from_part = html.element("from-prefix row gap", from_field)

    import_statement = html.element("importfrom row gap", from_part, import_part)
    yield from html.node(node, import_statement)




@register_node
def render_nonlocal(node: ast.Nonlocal):
    # TODO: test on multiple names (ex: nonlocal a, b, c)
    names = [text(name) for name in node.names]
    names = html.items("row comma-sep", "row gap", names)
    yield from element("row gap nonlocal-prefix", names)


def render_pass(node: ast.Pass):
    yield from html.node(node, text("pass"))


def render_continue(node: ast.Continue):
    yield from html.node(node, text("continue"))

